/* 
   $Id: nnmCanvas.c,v 1.11 2001/07/30 07:22:03 sho Exp $

   Copyright (C) 2001 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of Nunome.

   Nunome is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Nunome is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Nunome; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *	NuNoMe Project - 
 *		Canvas of Drawing with mouse
 *
 */
#include "config.h"
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Simple.h>
#include <X11/Xaw/Form.h>

#include "./nnmUI.h"
#include "nnmCanvas.h"
#include "nnmStub.h"

static	int		CurrentCanvas = -1;
static	NnmCanvasRec	canvases[MAX_NUM_CANVAS];
static int recog_top = 0;
static int recog_bottom = 0;
static int recog_canvas[MAX_NUM_CANVAS+1];


/*
 *  Windowɽ롼
 *
 *    ڥΥץ󥰥ǡ褷Ƥ뤬
 *  ɽˤϺƥץ󥰸Υǡɽ롣
 *    ʤʤ顢ȥ(Btn up)˥ȥ
 *  Ԥ뤿Ǥ롣
 */
static void	NnmCanvasExposure( Widget w, XtPointer clo,
					   XEvent *ev, Boolean *ctd )
{
  Display		*dpy = XtDisplay(w);
  Window		win  = XtWindow(w);
  int			i, n;
  GC			gc;
  int			idx = (int)clo;
  NnmCanvasRec		*canvasInfo;

  canvasInfo = &(canvases[idx]);
  if (canvasInfo->clean) return;
  if ( (gc = canvasInfo->gc)==NULL ){
    gc = canvasInfo->gc = XtGetGC( w, 0, 0 );
  }
  for( i=0; i<canvasInfo->CharInfo.num; i++ ){
    for( n=1; n<canvasInfo->CharInfo.stroke[i].len ; n++){
      int	x1,x2,y1,y2;
      
      x1 = canvasInfo->CharInfo.stroke[i].sample_x[n-1];
      x1 = x1*NnmCANVAS_DISPLAY_SIZE;
      x1 /= NnmCANVAS_INNER_SIZE;
      x2 = canvasInfo->CharInfo.stroke[i].sample_x[n  ];
      x2 = x2*NnmCANVAS_DISPLAY_SIZE;
      x2 /= NnmCANVAS_INNER_SIZE;
      y1 = canvasInfo->CharInfo.stroke[i].sample_y[n-1];
      y1 = y1*NnmCANVAS_DISPLAY_SIZE;
      y1 /= NnmCANVAS_INNER_SIZE;
      y2 = canvasInfo->CharInfo.stroke[i].sample_y[n  ];
      y2 = y2*NnmCANVAS_DISPLAY_SIZE;
      y2 /= NnmCANVAS_INNER_SIZE;
      XDrawLine( dpy, win, gc, x1, y1, x2, y2 );
    }
  }
  return ;
}

static int canvasIsBusy(int idx)
{
    int i;

    for (i = recog_top;i < recog_bottom;i++) {
	if (recog_canvas[i] == idx) return 1;
    }
    return 0;
}

extern int  recogTimeout;
static int  timerId = 0;

void NnmRemoveTimeOut()
{
  if(timerId){
    XtRemoveTimeOut(timerId);
  }
}

static void NnmCanvasTimerProc(caddr_t client_data, XtIntervalId id)
{
#ifdef WNN
  extern int IsConv; 
  if(IsConv){
    RevConv();
  }
#endif /* WNN */
  NnmCurrentCanvasRecognize( -1, 1 );
  timerId = 0;
}

/*
 *  ڥϳ: Btn1 Down
 */
static void	canvasDrawStart( Widget w, XtPointer clo,
				 XEvent *ev, Boolean *ctd )
{
  Display		*dpy = XtDisplay(w);
  Window		win  = XtWindow(w);
  int			x, y, num;
  GC			gc;
  NnmRawStrokeRec	*stroke;
  int			idx = (int)clo;
  NnmCanvasRec		*canvasInfo;

  NnmRemoveTimeOut();

  if (canvasIsBusy(idx)) return;
  if ( CurrentCanvas != idx )
    NnmCurrentCanvasRecognize( idx, 0 );
  canvasInfo = &(canvases[idx]);

  /* Button1 DownΤͭ */
  if ( ev->type != ButtonPress ) return;
  if ( ev->xbutton.button != Button1 ) return ;

  x  = ev->xbutton.x;
  y  = ev->xbutton.y;
  if ( (gc = canvasInfo->gc)==NULL ){	/* GCʤGC */
    gc = canvasInfo->gc = XtGetGC( w, 0, 0 );
  }
  XDrawPoint( dpy, win, gc, x, y );
  if ( canvasInfo->clean ) {
    canvasInfo->clean = False;
  }
  num = canvasInfo->CharInfo.num;
  stroke = &(canvasInfo->CharInfo.stroke[num]);
  stroke->sample_x[0] = x;
  stroke->sample_y[0] = y;
  stroke->len = 1;
  return ;
}



/*
 *  ڥ: Btn1 down & Mouse Dragging
 */
static void	canvasDrawing( Widget w, XtPointer clo,
				       XEvent *ev, Boolean *ctd )
{
  Display		*dpy = XtDisplay(w);
  Window		win  = XtWindow(w);
  int			x, y, oldX, oldY;
  GC			gc;
  NnmRawStrokeRec	*stroke;
  int			num, len;
  int			idx = (int)clo;
  NnmCanvasRec		*canvasInfo;

  if (canvasIsBusy(idx)) return;
  canvasInfo = &(canvases[idx]);

  if (canvasInfo->clean) {
    /* Motion without ButtonPress */
    /* We simulate ButtonPressEvent */
    XEvent newev;

    newev.type = ButtonPress;
    newev.xbutton.button = Button1;
    newev.xbutton.x = ev->xmotion.x;
    newev.xbutton.y = ev->xmotion.y;
    return canvasDrawStart(w,clo,&newev,ctd);
  }

  if ( ev->type != MotionNotify ) return;
  x  = ev->xmotion.x;
  y  = ev->xmotion.y;
  /* ɥ */
  if ( (x<0)||(x>=NnmCANVAS_DISPLAY_SIZE)||
       (y<0)||(y>=NnmCANVAS_DISPLAY_SIZE) ){
    return;
  }
  num = canvasInfo->CharInfo.num;
  stroke = &(canvasInfo->CharInfo.stroke[num]);
  len = MAX( stroke->len-1, 0 );
  oldX = stroke->sample_x[len];
  oldY = stroke->sample_y[len];
  if ( stroke->len < NnmMaxSTROKE_LEN-1 ){
    gc = canvasInfo->gc;
    XDrawLine( dpy, win, gc, oldX, oldY, x, y );
    stroke->sample_x[stroke->len] = x;
    stroke->sample_y[stroke->len] = y;
    stroke->len++;
  }
}

/*
 *  ڥϽλ: Btn1 up
 */
static void	canvasDrawEnd( Widget w, XtPointer clo,
			       XEvent *ev, Boolean *ctd )
{
  int			idx = (int)clo;
  Display		*dpy = XtDisplay(w);
  Window		win  = XtWindow(w);
  int			x, y, oldX, oldY, num, len;
  GC			gc;
  NnmRawStrokeRec	*stroke;
  NnmCanvasRec		*canvasInfo;

  if (canvasIsBusy(idx)) return;
  canvasInfo = &(canvases[idx]);

  if (canvasInfo->clean) {
    /* ButtonRelease without ButtonPress */
    /* We simulate ButtonPressEvent */

    canvasDrawStart(w,clo,ev,ctd);
  }

  if ( ev->type != ButtonRelease ) return;
  if ( ev->xbutton.button != Button1 ) return ;
  x  = ev->xbutton.x;
  y  = ev->xbutton.y;
  /* ɥ */
  if ( x >= 0 && x < NnmCANVAS_DISPLAY_SIZE &&
       y >= 0 && y < NnmCANVAS_DISPLAY_SIZE ){
    num = canvasInfo->CharInfo.num;
    stroke = &(canvasInfo->CharInfo.stroke[num]);
    len = stroke->len - 1;
    len = MAX( 0, len );
    oldX = stroke->sample_x[len];
    oldY = stroke->sample_y[len];
    if ( stroke->len < NnmMaxSTROKE_LEN-1 ){
      gc = canvasInfo->gc;
      XDrawLine( dpy, win, gc, oldX, oldY, x, y );
      stroke->sample_x[stroke->len] = x;
      stroke->sample_y[stroke->len] = y;
      stroke->len++;
    }
  }
  if ( canvasInfo->CharInfo.num < NnmMaxSTROKE_NUM-1 ){
    canvasInfo->CharInfo.num++;		/* ȥ */
  }

  timerId = XtAppAddTimeOut(XtDisplayToApplicationContext(XtDisplay(w)),
			    recogTimeout, 
			    (XtTimerCallbackProc)NnmCanvasTimerProc, NULL);
  
}


/*
 *  ڥѥХɥκ
 *
 *
 */
static void	NnmCreateCanvas( NnmCanvasRec *canvasInfo, int canvasNo, 
				 Widget parent, Arg args[], int num_args )
{
  Widget	w;

  /* set Window Size */
  canvasInfo->canvas = w = 
    XtCreateManagedWidget ( "canvas", simpleWidgetClass,
			    parent, args, num_args);

  /* Redraw on Exposure Event */
  /*  */
  XtAddEventHandler( w, ExposureMask, (Boolean)0, 
		     (XtEventHandler)NnmCanvasExposure, (XtPointer)canvasNo );
  /* Button1 Down */
  XtAddEventHandler( w, ButtonPressMask, (Boolean)0, 
		     (XtEventHandler)canvasDrawStart, (XtPointer)canvasNo );
  /* Button Motion */
  XtAddEventHandler( w, Button1MotionMask, (Boolean)0, 
		     (XtEventHandler)canvasDrawing, (XtPointer)canvasNo );
  /* Button1 Up */
  XtAddEventHandler( w, ButtonReleaseMask, (Boolean)0, 
		     (XtEventHandler)canvasDrawEnd, (XtPointer)canvasNo );

  canvasInfo->gc = NULL;	/* GC: ǽΥڥϻ˽Ϥ */
  canvasInfo->clean = True;
  canvasInfo->CharInfo.num = 0; /*  */
}


/*
 *  ڥѥХɥʣ
 */
void	NnmCreateCanvases( Widget parent )
{
  int		i, j;
  Arg		args[10];

  i = 0;
  XtSetArg( args[i], XtNfromHoriz,  NULL ); i++;
  XtSetArg( args[i], XtNfromVert,   NULL ); i++;
  XtSetArg( args[i], XtNvertDistance, 0 ); i++;
  XtSetArg( args[i], XtNhorizDistance, 0 ); i++;
  XtSetArg( args[i], XtNwidth,  NnmCANVAS_DISPLAY_SIZE ); i++;
  XtSetArg( args[i], XtNheight, NnmCANVAS_DISPLAY_SIZE ); i++;
  for( j=0; j<MAX_NUM_CANVAS; j++ ){
    NnmCreateCanvas( &(canvases[j]), j, parent, args, i );
    XtSetArg( args[0], XtNfromHoriz, canvases[j].canvas );
  }
}


int	NnmCanvasClear(int idx)
{
  Display	*dpy;
  Window	w;
  int		ret;

  ret = idx;
  if ( idx < 0 ) return ret;
  canvases[idx].CharInfo.num = 0;
  canvases[idx].clean = True;
  dpy = XtDisplay( canvases[idx].canvas );
  w   = XtWindow(  canvases[idx].canvas );
  XClearWindow( dpy, w );
  return ret;
}


/*
 *  Window Clear롼
 *
 *    ߥڥCanvas Window򥯥ꥢ롣
 *    Ʊ˥ȥ⥯ꥢ롣
 */
int	NnmCurrentCanvasClear(void)
{
  int ret;

  ret = NnmCanvasClear(CurrentCanvas);
  CurrentCanvas = -1;
  return ret;
}


void NnmHandleRecognizeResult()
{
    ulong	result_code;
    ulong	cands[MAX_CANDIDATE_CATEGORY];
    int idx;
    NnmFeatureRec fv;

    result_code = NnmStubResultRecognizeCharInfo(cands, MAX_CANDIDATE_CATEGORY, &fv);
    idx = recog_canvas[recog_top++];
    if (recog_top > MAX_NUM_CANVAS) recog_top = 0;
    NnmAppendResult( result_code,
		   &(canvases[idx].CharInfo),
		   cands,&fv);
    NnmCanvasClear(idx);
}

/*
 *  ǧ¹
 *
 *
 */
int	NnmCurrentCanvasRecognize( int NextCanvas, int block )
{
  if ( CurrentCanvas==-1 ){
    CurrentCanvas = NextCanvas;
    return 0;
  }
  if (canvases[CurrentCanvas].clean || canvases[CurrentCanvas].CharInfo.num == 0) {
    CurrentCanvas = NextCanvas;
    return 0;
  }
  NnmStubRecognizeCharInfo( &(canvases[CurrentCanvas].CharInfo),
			  MAX_CANDIDATE_CATEGORY );
  recog_canvas[recog_bottom++] = CurrentCanvas;
  if (recog_bottom > MAX_NUM_CANVAS) recog_bottom = 0;
  if (block) {
    while (recog_top != recog_bottom) {
	NnmHandleRecognizeResult();
    }
  }
  CurrentCanvas = NextCanvas;
  return 0;
}


